@@ -23,7 +23,7 @@
  *
  * Author: Bert Bos <bert@w3.org>
  * Created Sep 1997
- * Version: $Id: hxtoc.c,v 1.12 2017/11/24 09:50:25 bbos Exp $
+ * Version: $Id: hxtoc.c,v 1.13 2019/08/28 19:14:07 bbos Exp $
  *
  **/
 #include "config.h"
@@ -34,6 +34,7 @@
 #include <time.h>
 #include <stdbool.h>
 #include <getopt.h>
+#include <limits.h>
 #if STDC_HEADERS
 # include <string.h>
 #else
@@ -83,7 +84,7 @@
 #define INDENT " "				/* Amount to indent ToC per level */
 
 static Tree tree;
-static int toc_low = 1, toc_high = 6;		/* Which headers to include */
+static int toc_low = 1, toc_high = INT_MAX;	/* Which headers to include */
 static bool xml = false;			/* Use <empty /> convention */
 static bool bctarget = true;			/* Generate <a name=> after IDs */
 static string toc_class = "toc";		/* <ul class="..."> */
@@ -103,7 +104,7 @@ static void* start(void)
   tree = create();
   return NULL;
 }
-  
+
 /* end -- called after the last event is reported */
 static void end(void *clientdata)
 {
@@ -189,35 +190,38 @@ static int heading_level(Tree t)
   return 0;
 }
 
-/* div_parent -- if t is the first child of a section elt, return that elt */
+/* div_parent -- if t is the first heading in a section elt, return that elt */
 static Tree div_parent(Tree t)
 {
   Tree h, result = NULL;
 
   assert(t->tp == Element);
   assert(t->parent);
+  assert(eq(t->name, "hgroup") || heading_level(t) > 0);
   if (t->parent->tp != Element) return NULL;
   if (has_class(t->parent->attribs, NO_TOC)) return NULL;
   if (is_div(t->parent)) result = t->parent;
-  else if (!eq(t->parent->name, "header")) return NULL;
+  else if (!eq(t->parent->name, "hgroup")) return NULL;
   else if (!(result = div_parent(t->parent))) return NULL;
-  for (h = t->parent->children; h != t; h = h->sister) {
-    if (h->tp == Element) return NULL;
-    if (h->tp == Text && !only_space(h->text)) return NULL;
-  }
-  return result;
+
+  /* Check if t is the first heading in its parent. */
+  for (h = t->parent->children; h != t; h = h->sister)
+    if (h->tp == Element && (eq(h->name, "hgroup") || heading_level(h) > 0))
+      return NULL;		/* No, it's not */
+  return result;		/* Yes, it is */
 }
 
-/* first_child_is_heading -- true if first child is a Hn or HEADER */
-static bool first_child_is_heading(Tree t)
+/* has_heading -- true if the element has at least one Hn or HGROUP as child */
+static bool has_heading(Tree t)
 {
   Tree h;
 
   assert(t->tp == Element);
   for (h = t->children; h; h = h->sister) {
+#if 0
     switch (h->tp) {
     case Element:
-      return eq(h->name, "header") || heading_level(h) > 0;
+      return eq(h->name, "hgroup") || heading_level(h) > 0;
     case Text:
       if (!only_space(h->text))
 	return false;
@@ -225,6 +229,10 @@ static bool first_child_is_heading(Tree t)
     default:
       break;
     }
+#else
+    if (h->tp == Element &&
+	(eq(h->name, "hgroup") || heading_level(h) > 0)) return true;
+#endif
   }
   return false;
 }
@@ -246,8 +254,8 @@ static void toc(Tree t, int *curlevel, bool *item_is_open, int div_depth)
     case Declaration: break;
     case Procins: break;
     case Element:
-      if (use_div && is_div(t) && first_child_is_heading(t)) {
-	/* It's a section element with a heading as first child */
+      if (use_div && is_div(t) && has_heading(t)) {
+	/* It's a section element with a heading */
 	div_depth++;
 	level = 0;
       } else {
@@ -358,7 +366,7 @@ static void expand(Tree t, bool *write, bool exp, bool keep_anchors,
 	if (*write) printf("<?%s>", h->text);
 	break;
       case Element:
-	if (use_div && is_div(h) && first_child_is_heading(h)) {
+	if (use_div && is_div(h) && has_heading(h)) {
 	  /* It's a section element with a heading as first child */
 	  div_depth++;
 	  level = div_depth;
@@ -457,7 +465,6 @@ int main(int argc, char *argv[])
     }
   }
   if (toc_low < 1) toc_low = 1;
-  if (toc_high > 6) toc_high = 6;
 
   if (argc > optind + 1) {
     usage(argv[0]);
