@@ -6,7 +6,7 @@
  *
  * Bert Bos <bert@w3.org>
  * Created 31 July 1999
- * $Id: hxwls.c,v 1.12 2017/12/07 02:05:23 bbos Exp $
+ * $Id: hxwls.c,v 1.13 2019/08/28 19:14:34 bbos Exp $
  */
 #include "config.h"
 #include <assert.h>
@@ -172,7 +172,7 @@ void* start(void)
   }
   return NULL;
 }
-  
+
 /* end -- called after the last event is reported */
 void end(void *clientdata)
 {
@@ -228,13 +228,14 @@ void handle_starttag(void *clientdata, string name, pairlist attribs)
     output("a", pairlist_get(attribs, "rel"), pairlist_get(attribs, "href"));
   } else if (strcasecmp(name, "img") == 0) {
     output("img", NULL, pairlist_get(attribs, "src"));
-    output("longdesc", NULL, pairlist_get(attribs, "longdesc"));
+    output("img", "longdesc", pairlist_get(attribs, "longdesc"));
+    output("img", "srcset", pairlist_get(attribs, "srcset"));
   } else if (strcasecmp(name, "input") == 0) {
-    output("input", NULL, pairlist_get(attribs, "src"));
+    output("input", "src", pairlist_get(attribs, "src"));
   } else if (strcasecmp(name, "object") == 0) {
     output("object", NULL,  pairlist_get(attribs, "data"));
-    output("object", NULL,  pairlist_get(attribs, "classid"));
-    output("object", NULL,  pairlist_get(attribs, "codebase"));
+    output("object", "classid",  pairlist_get(attribs, "classid"));
+    output("object", "codebase",  pairlist_get(attribs, "codebase"));
   } else if (strcasecmp(name, "area") == 0) {
     output("area", pairlist_get(attribs, "rel"), pairlist_get(attribs, "href"));
   } else if (strcasecmp(name, "ins") == 0) {
@@ -262,8 +263,8 @@ void handle_starttag(void *clientdata, string name, pairlist attribs)
   } else if (strcasecmp(name, "audio") == 0) {
     output("audio", NULL, pairlist_get(attribs, "src"));
   } else if (strcasecmp(name, "source") == 0) {
-    output("source", NULL, pairlist_get(attribs, "srcset"));
-    output("source", NULL, pairlist_get(attribs, "src"));
+    output("source", "srcset", pairlist_get(attribs, "srcset"));
+    output("source", "src", pairlist_get(attribs, "src"));
   }
 
   /* Free memory */
