@@ -8,7 +8,7 @@
  *
  * Bert Bos
  * Created 10 May 1998
- * $Id: textwrap.c,v 1.32 2017/11/24 10:14:49 bbos Exp $
+ * $Id: textwrap.c,v 1.34 2019/10/05 17:49:44 bbos Exp $
  */
 #include "config.h"
 #include <stdio.h>
@@ -63,7 +63,7 @@ EXPORT void flush()
       else if ((buf[j] & 0xC0) != 0x80) k++; /* Start of a UTF-8 sequence */
     if (i < 0) break;				/* No breakpoint */
     assert(i >= 0);				/* Found a breakpoint at i */
-    assert(buf[i] == ' ' || buf[i]==BREAKOP);
+    assert(buf[i] == ' ' || buf[i] == BREAKOP);
     /* Print up to breakpoint (removing non-break-space markers) */
     for (j = 0; j < i; j++)
       if (buf[j] != BREAKOP) {
@@ -98,17 +98,24 @@ EXPORT void flush()
 }
 
 /* outc -- add one character to output buffer */
-EXPORT void outc(char c, bool preformatted)
+EXPORT void outc(char c, bool preformatted, bool with_space)
 {
-  if (c == '\n' && !preformatted) c = ' ';	/* Newline is just a space */
-  if (c == '\r' && !preformatted) c = ' ';	/* CR is just a space */
-  if (c == '\t' && !preformatted) c = ' ';	/* Tab is just a space */
-  if (c == '\f' && !preformatted) c = ' ';	/* Formfeed is just a space */
-  if (c == ' ' && preformatted) c = NBSP;	/* Non-break-space marker */
-  if (c == ' ' && prev == ' ') return;		/* Don't add another space */
+  if (c == '\n' || c == '\r' || c == '\f') {
+    if (preformatted) ;		  /* Keep unchanged */
+    else if (with_space) c = ' '; /* Treated as space */
+    else c = BREAKOP;		  /* Treated as a break opportunity */
+  } else if (c == '\t') {
+    if (preformatted) ;		  /* Keep unchanged */
+    else c = ' ';		  /* Tab is just a space */
+  }
+  if (c == ' ') {
+    if (preformatted) c = NBSP;	  /* Non-break-space marker */
+    else if (prev == ' ') return; /* Don't add another space */
+    else if (prev == BREAKOP) return; /* Don't add a space after \n or similar */
+  }
   if ((c == ' ' || c == BREAKOP) && linelen + bufchars >= maxlinelen) flush();
   if (c == '\n' || c == '\r' || c == '\f') flush(); /* Empty the buf */
-  if (c == ' ' && linelen + len == 0) return;	/* No ins at BOL */
+  if (c == ' ' && linelen + len == 0) return;	/* No insert at BOL */
   while (level * indent >= buflen) {buflen += 1024; renewarray(buf, buflen);}
   if (linelen + len == 0 && !preformatted)
     while (len < level * indent) {buf[len++] = NBSP; bufchars++;}
@@ -120,22 +127,22 @@ EXPORT void outc(char c, bool preformatted)
 }
 
 /* out -- add text to current output line, print line if getting too long */
-EXPORT void out(string s, bool preformatted)
+EXPORT void out(string s, bool preformatted, bool with_space)
 {
-  if (s) for (; *s; s++) outc(*s, preformatted);
+  if (s) for (; *s; s++) outc(*s, preformatted, with_space);
 }
 
 /* outn -- add n chars to current output, print line if getting too long */
-EXPORT void outn(string s, size_t n, bool preformatted)
+EXPORT void outn(string s, size_t n, bool preformatted, bool with_space)
 {
   size_t i;
-  for (i = 0; i < n; i++) outc(s[i], preformatted);
+  for (i = 0; i < n; i++) outc(s[i], preformatted, with_space);
 }
 
 /* outln -- add string to output buffer, followed by '\n' */
-EXPORT void outln(char *s, bool preformatted)
+EXPORT void outln(char *s, bool preformatted, bool with_space)
 {
-  out(s, preformatted);
+  out(s, preformatted, with_space);
   flush();
   assert(len == 0);
   assert(bufchars == 0);
@@ -158,7 +165,7 @@ EXPORT void outbreak(void)
 /* outbreakpoint -- mark a possible line break point */
 EXPORT void outbreakpoint(void)
 {
-  outc(BREAKOP, false);
+  outc(BREAKOP, false, true);
 }
 
 
